"""rio_cogeo.utils: Utility functions."""

from typing import Dict, Optional, Tuple, Union

import morecantile
from rasterio.enums import ColorInterp, MaskFlags
from rasterio.io import DatasetReader, DatasetWriter
from rasterio.rio.overview import get_maximum_overview_level
from rasterio.transform import Affine
from rasterio.vrt import WarpedVRT
from rasterio.warp import calculate_default_transform


def has_alpha_band(src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT]):
    """Check for alpha band or mask in source."""
    if (
        any([MaskFlags.alpha in flags for flags in src_dst.mask_flag_enums])
        or ColorInterp.alpha in src_dst.colorinterp
    ):
        return True
    return False


def has_mask_band(src_dst):
    """Check for mask band in source."""
    if any(
        [
            (MaskFlags.per_dataset in flags and MaskFlags.alpha not in flags)
            for flags in src_dst.mask_flag_enums
        ]
    ):
        return True
    return False


def non_alpha_indexes(src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT]) -> Tuple:
    """Return indexes of non-alpha bands."""
    return tuple(
        b
        for ix, b in enumerate(src_dst.indexes)
        if (
            src_dst.mask_flag_enums[ix] is not MaskFlags.alpha
            and src_dst.colorinterp[ix] is not ColorInterp.alpha
        )
    )


def get_zooms(
    src_dst: Union[DatasetReader, DatasetWriter, WarpedVRT],
    tilesize: int = 256,
    tms: morecantile.TileMatrixSet = morecantile.tms.get("WebMercatorQuad"),
    zoom_level_strategy: str = "auto",
) -> Tuple[int, int]:
    """Calculate raster min/max zoom level."""
    # If the raster is not in the TMS CRS we calculate its projected properties (height, width, resolution)
    tms_crs = tms.rasterio_crs
    if src_dst.crs != tms_crs:
        aff, w, h = calculate_default_transform(
            src_dst.crs,
            tms_crs,
            src_dst.width,
            src_dst.height,
            *src_dst.bounds,
        )
    else:
        aff = list(src_dst.transform)
        w = src_dst.width
        h = src_dst.height

    resolution = max(abs(aff[0]), abs(aff[4]))

    # The maxzoom is defined by finding the minimum difference between
    # the raster resolution and the zoom level resolution
    max_zoom = tms.zoom_for_res(
        resolution,
        max_z=30,
        zoom_level_strategy=zoom_level_strategy,
    )

    # The minzoom is defined by the resolution of the maximum theoretical overview level
    max_possible_overview_level = get_maximum_overview_level(w, h, minsize=tilesize)
    ovr_resolution = resolution * (2**max_possible_overview_level)
    min_zoom = tms.zoom_for_res(ovr_resolution, max_z=30)

    return (min_zoom, max_zoom)


def get_web_optimized_params(
    src_dst,
    zoom_level_strategy: str = "auto",
    zoom_level: Optional[int] = None,
    aligned_levels: Optional[int] = None,
    tms: morecantile.TileMatrixSet = morecantile.tms.get("WebMercatorQuad"),
) -> Dict:
    """Return VRT parameters for a WebOptimized COG."""
    tms_crs = tms.rasterio_crs

    if src_dst.crs != tms_crs:
        with WarpedVRT(src_dst, crs=tms_crs) as vrt:
            bounds = vrt.bounds
            aff = list(vrt.transform)

    else:
        bounds = src_dst.bounds
        aff = list(src_dst.transform)

    resolution = max(abs(aff[0]), abs(aff[4]))

    if zoom_level is None:
        # find max zoom (closest to the raster resolution)
        max_zoom = tms.zoom_for_res(
            resolution,
            max_z=30,
            zoom_level_strategy=zoom_level_strategy,
        )

    else:
        max_zoom = zoom_level

    # defined the zoom level we want to align the raster
    aligned_levels = aligned_levels or 0
    base_zoom = max_zoom - aligned_levels

    # find new raster bounds (bounds of UL tile / LR tile)
    ul_tile = tms._tile(bounds[0], bounds[3], base_zoom)
    w, _, _, n = tms.xy_bounds(ul_tile)

    # The output resolution should match the TMS resolution at MaxZoom
    vrt_res = tms.matrix(max_zoom).cellSize

    # Output transform is built from the origin (UL tile) and output resolution
    vrt_transform = Affine(vrt_res, 0, w, 0, -vrt_res, n)

    lr_tile = tms._tile(bounds[2], bounds[1], base_zoom)
    e, _, _, s = tms.xy_bounds(
        morecantile.Tile(lr_tile.x + 1, lr_tile.y + 1, lr_tile.z)
    )

    vrt_width = max(1, round((e - w) / vrt_transform.a))
    vrt_height = max(1, round((s - n) / vrt_transform.e))

    return {
        "crs": tms_crs,
        "transform": vrt_transform,
        "width": vrt_width,
        "height": vrt_height,
    }



"""Rio-Cogeo Errors and Warnings."""


class LossyCompression(UserWarning):
    """Rio-cogeo module Lossy compression warning."""


class IncompatibleBlockRasterSize(UserWarning):
    """Rio-cogeo module incompatible raster block/size warning."""


class NodataAlphaMaskWarning(UserWarning):
    """Dataset with nodata and alpha/mask warning."""


class RioCogeoError(Exception):
    """Base exception class."""


class IncompatibleOptions(RioCogeoError):
    """Rio-cogeo module incompatible options."""



"""rio-cogeo models."""

from typing import Any, Dict, Optional, Sequence, Tuple

from pydantic import BaseModel, Field

BBox = Tuple[float, float, float, float]


class DictLike:
    """Provides dictionary access for pydantic models, for backwards compatability with rio-cogeo<2.2.0."""

    def __getitem__(self, item):
        """Access item like in Dict."""
        return self.__dict__[item]


class IFD(DictLike, BaseModel):
    """ImageFileDirectory info."""

    Level: int
    Width: int
    Height: int
    Blocksize: Tuple[int, int]
    Decimation: int


class Geo(DictLike, BaseModel):
    """rio-cogeo validation GEO information."""

    CRS: Optional[str] = None
    BoundingBox: BBox
    Origin: Tuple[float, float]
    Resolution: Tuple[float, float]
    MinZoom: Optional[int] = None
    MaxZoom: Optional[int] = None


class Profile(DictLike, BaseModel):
    """rio-cogeo validation Profile information."""

    Bands: int
    Width: int
    Height: int
    Tiled: bool
    Dtype: str
    Interleave: str
    AlphaBand: bool
    InternalMask: bool
    Nodata: Any = None
    ColorInterp: Sequence[str]
    ColorMap: bool
    Scales: Sequence[float]
    Offsets: Sequence[float]

    model_config = {"extra": "ignore"}


class BandMetadata(DictLike, BaseModel):
    """Band metadata."""

    Description: Optional[str] = None
    ColorInterp: str
    Offset: float
    Scale: float
    Metadata: Optional[Dict[str, Any]] = None


class Info(DictLike, BaseModel):
    """rio-cogeo Info."""

    Path: str
    Driver: str
    COG: bool
    Compression: Optional[str] = None
    ColorSpace: Optional[str] = None
    COG_errors: Optional[Sequence[str]] = None
    COG_warnings: Optional[Sequence[str]] = None
    Profile: Profile
    GEO: Geo
    Tags: Dict[str, Dict]
    Band_Metadata: Dict[str, BandMetadata] = Field(None, alias="Band Metadata")
    IFD: Sequence[IFD]

    model_config = {"populate_by_name": True}



"""rio_cogeo.cogeo: translate a file to a cloud optimized geotiff."""

import os
import pathlib
import sys
import tempfile
import warnings
from contextlib import ExitStack, contextmanager
from typing import Any, Dict, List, Literal, Optional, Sequence, TextIO, Tuple, Union

import click
import morecantile
import rasterio
from rasterio.enums import ColorInterp
from rasterio.enums import Resampling as ResamplingEnums
from rasterio.env import GDALVersion
from rasterio.io import DatasetReader, DatasetWriter, MemoryFile
from rasterio.rio.overview import get_maximum_overview_level
from rasterio.shutil import copy
from rasterio.transform import from_gcps as transform_from_gcps
from rasterio.vrt import WarpedVRT

from rio_cogeo import models, utils
from rio_cogeo.errors import IncompatibleOptions, NodataAlphaMaskWarning

IN_MEMORY_THRESHOLD = int(os.environ.get("IN_MEMORY_THRESHOLD", 10980 * 10980))


@contextmanager
def TemporaryRasterFile(dst_path: Union[str, pathlib.PurePath], suffix: str = ".tif"):
    """Create temporary file."""
    # For local file we should create temporary file in the same directory
    tmpdir = (
        pathlib.Path(dst_path).parent
        if pathlib.Path(dst_path).parent.is_dir()
        else None
    )
    fileobj = tempfile.NamedTemporaryFile(dir=tmpdir, suffix=suffix, delete=False)
    fileobj.close()
    try:
        yield fileobj
    finally:
        os.remove(fileobj.name)


# RasterIO() resampling method.
# ref: https://gdal.org/api/raster_c_api.html#_CPPv418GDALRIOResampleAlg
RIOResampling = Literal[
    "nearest",
    "bilinear",
    "cubic",
    "cubic_spline",
    "lanczos",
    "average",
    "mode",
    "gauss",
    "rms",
]

# WarpKernel resampling method.
# ref: https://gdal.org/api/gdalwarp_cpp.html#_CPPv4N14GDALWarpKernel9eResampleE
WarpResampling = Literal[
    "nearest",
    "bilinear",
    "cubic",
    "cubic_spline",
    "lanczos",
    "average",
    "mode",
    "sum",
    "rms",
]


def cog_translate(  # noqa: C901
    source: Union[str, pathlib.PurePath, DatasetReader, DatasetWriter, WarpedVRT],
    dst_path: Union[str, pathlib.PurePath],
    dst_kwargs: Dict,
    indexes: Optional[Sequence[int]] = None,
    nodata: Optional[Union[str, int, float]] = None,
    dtype: Optional[str] = None,
    add_mask: bool = False,
    overview_level: Optional[int] = None,
    overview_resampling: RIOResampling = "nearest",
    web_optimized: bool = False,
    tms: Optional[morecantile.TileMatrixSet] = None,
    zoom_level_strategy: str = "auto",
    zoom_level: Optional[int] = None,
    aligned_levels: Optional[int] = None,
    resampling: WarpResampling = "nearest",
    in_memory: Optional[bool] = None,
    config: Optional[Dict] = None,
    allow_intermediate_compression: bool = False,
    forward_band_tags: bool = False,
    forward_ns_tags: bool = False,
    quiet: bool = False,
    progress_out: Optional[TextIO] = None,
    temporary_compression: str = "DEFLATE",
    colormap: Optional[Dict] = None,
    additional_cog_metadata: Optional[Dict] = None,
    use_cog_driver: bool = False,
    decimation_base: int = 2,
):
    """
    Create Cloud Optimized Geotiff.

    Parameters
    ----------
    source : str, PathLike object or rasterio.io.DatasetReader
        A dataset path, URL or rasterio.io.DatasetReader object.
        Will be opened in "r" mode.
    dst_path : str or PathLike object
        An output dataset path or or PathLike object.
        Will be opened in "w" mode.
    dst_kwargs: dict
        Output dataset creation options.
    indexes : tuple or int, optional
        Raster band indexes to copy.
    nodata, int, optional
        Overwrite nodata masking values for input dataset.
    dtype: str, optional
        Overwrite output data type. Default will be the input data type.
    add_mask, bool, optional
        Force output dataset creation with a mask.
    overview_level : int, optional (default: None)
        COGEO overview (decimation) level. By default, inferred from data size.
    overview_resampling : str, optional (default: "nearest")
        RasterIO Resampling algorithm for overviews
    web_optimized: bool, optional (default: False)
        Create web-optimized cogeo.
        Deprecated: Behavior changed since 5.1.2. See deprecation warning at the bottom of function doc.
    tms: morecantile.TileMatrixSet, optional
        TileMatrixSet to use for reprojection, resolution and alignment.
    zoom_level_strategy: str, optional (default: auto)
        Strategy to determine zoom level (same as in GDAL 3.2).
        Used only when `tms` is not None.
        LOWER will select the zoom level immediately below the theoretical computed non-integral zoom level, leading to subsampling.
        On the contrary, UPPER will select the immediately above zoom level, leading to oversampling.
        Defaults to AUTO which selects the closest zoom level.
        ref: https://gdal.org/drivers/raster/cog.html#raster-cog
    zoom_level: int, optional.
        Zoom level number (starting at 0 for coarsest zoom level).
        If this option is specified, `--zoom-level-strategy` is ignored.
        In any case, it is used only when `tms` is not None.
    aligned_levels: int, optional.
        Number of overview levels for which GeoTIFF tile and tiles defined in the tiling scheme match.
        Used only when `tms` is not None.
        Default is to use the maximum overview levels. Note: GDAL use number of resolution levels instead of overview levels.
    resampling : str, optional (default: "nearest")
        Warp Resampling algorithm.
    in_memory: bool, optional
        Force processing raster in memory (default: process in memory if small)
    config : dict
        Rasterio Env options.
    allow_intermediate_compression: bool, optional (default: False)
        Allow intermediate file compression to reduce memory/disk footprint.
        Note: This could reduce the speed of the process.
        Ref: https://github.com/cogeotiff/rio-cogeo/issues/103
    forward_band_tags:  bool, optional
        Forward band tags to output bands.
        Ref: https://github.com/cogeotiff/rio-cogeo/issues/19
    forward_ns_tags:  bool, optional
        Forward namespaces tags to output dataset.
    quiet: bool, optional (default: False)
        Mask processing steps.
    progress_out: TextIO, optional
        Output progress steps to alternative text buffer. Quiet must be False.
    temporary_compression: str, optional
        Compression used for the intermediate file, default is deflate.
    colormap: dict, optional
        Overwrite or add a colormap to the output COG.
    additional_cog_metadata: dict, optional
        Additional dataset metadata to add to the COG.
    use_cog_driver: bool, optional (default: False)
        Use GDAL COG driver if set to True. COG driver is available starting with GDAL 3.1.
    decimation_base: int, default: 2
        How overviews are divided at each zoom level (default is 2). Must be greater than 1.
        Also requires that `overview_level` is provided for `decimation_base` values greater
        than 2.

    .. deprecated:: 5.1.2
        `web_optimized` is deprecated in favor of `tms`.
        Previously, `tms` usage was conditioned by `web_optimized` state.
        The behaviour has changed to allow setting a tile matrix set without the need of `web_optimized` flag.
        `web_optimized` now only serve to activate a default `WebMercatorQuad` tile matrix set.
        Set to be removed 6.0.

    """
    if web_optimized:
        warnings.warn(
            "'web_optomized' option is deprecated and will be removed in 6.0. Please use the `tms` option",
            DeprecationWarning,
        )
        tms = tms or morecantile.tms.get("WebMercatorQuad")

    if decimation_base <= 1:
        raise ValueError(
            "Decimation base must be greater than 1 for building overviews."
        )
    elif decimation_base > 2 and overview_level is None:
        raise ValueError(
            "Decimation base values greater than 2 require that overview_level is defined."
        )

    dst_kwargs = dst_kwargs.copy()

    if isinstance(indexes, int):
        indexes = (indexes,)

    config = config or {}
    with rasterio.Env(**config):
        with ExitStack() as ctx:
            if isinstance(source, (DatasetReader, DatasetWriter, WarpedVRT)):
                src_dst = source
            else:
                src_dst = ctx.enter_context(rasterio.open(source))

            meta = src_dst.meta
            indexes = indexes if indexes else src_dst.indexes
            nodata = nodata if nodata is not None else src_dst.nodata
            dtype = dtype if dtype else src_dst.dtypes[0]
            alpha = utils.has_alpha_band(src_dst)
            mask = utils.has_mask_band(src_dst)

            if nodata is not None and (alpha or mask):
                warnings.warn(
                    "Input dataset has both a nodata value and internal alpha/mask band. Nodata value will be prioritized.",
                    NodataAlphaMaskWarning,
                )

            if colormap and len(indexes) > 1:
                raise IncompatibleOptions(
                    "Cannot add a colormap for multiple bands data."
                )

            # we need to remove the `alpha band` index from the output data
            # when we translate the band to an internal mask
            if alpha and add_mask:
                indexes = tuple(
                    x for x in indexes if x in utils.non_alpha_indexes(src_dst)
                )

            if not add_mask and (
                (nodata is not None or alpha)
                and dst_kwargs.get("compress", "").lower() == "jpeg"
            ):
                warnings.warn(
                    "Nodata/Alpha band will be translated to an internal mask band.",
                )
                add_mask = True
                indexes = (
                    utils.non_alpha_indexes(src_dst)
                    if len(indexes) not in [1, 3]
                    else indexes
                )

            src_indexes = indexes

            tilesize = min(int(dst_kwargs["blockxsize"]), int(dst_kwargs["blockysize"]))

            vrt_params = {
                "dtype": dtype,
                "width": src_dst.width,
                "height": src_dst.height,
                "resampling": ResamplingEnums[resampling],
            }
            if src_dst.gcps[1]:
                vrt_params.update(
                    {
                        "add_alpha": True,
                        "src_crs": src_dst.gcps[1],
                        "src_transform": transform_from_gcps(src_dst.gcps[0]),
                    }
                )

            if nodata is not None:
                vrt_params.update(
                    {"nodata": nodata, "add_alpha": False, "src_nodata": nodata}
                )

            elif alpha:
                vrt_params.update({"add_alpha": False})

            elif mask:
                vrt_params.update({"add_alpha": True})

            if tms:
                wo_params = utils.get_web_optimized_params(
                    src_dst,
                    zoom_level_strategy=zoom_level_strategy,
                    zoom_level=zoom_level,
                    aligned_levels=aligned_levels,
                    tms=tms,
                )
                vrt_params.update(**wo_params)

            if vrt_params.get("add_alpha", False) and not (add_mask or mask):
                indexes = tuple(indexes) + (src_dst.count + 1,)

            with WarpedVRT(src_dst, **vrt_params) as vrt_dst:
                meta = vrt_dst.meta
                meta["count"] = len(indexes)

                if add_mask:
                    meta.pop("nodata", None)
                    meta.pop("alpha", None)

                if (
                    dst_kwargs.get("photometric", "").upper() == "YCBCR"
                    and meta["count"] == 1
                ):
                    warnings.warn(
                        "PHOTOMETRIC=YCBCR not supported on a 1-band raster"
                        " and has been set to 'MINISBLACK'"
                    )
                    dst_kwargs["photometric"] = "MINISBLACK"

                meta.update(**dst_kwargs)
                meta.pop("compress", None)
                meta.pop("photometric", None)

                if allow_intermediate_compression:
                    meta["compress"] = temporary_compression

                if in_memory is None:
                    in_memory = vrt_dst.width * vrt_dst.height < IN_MEMORY_THRESHOLD

                if in_memory:
                    tmpfile = ctx.enter_context(MemoryFile())
                    tmp_dst = ctx.enter_context(tmpfile.open(**meta))
                else:
                    tmpfile = ctx.enter_context(TemporaryRasterFile(dst_path))
                    tmp_dst = ctx.enter_context(
                        rasterio.open(tmpfile.name, "w", **meta)
                    )

                # Transfer color interpolation
                if len(indexes) == 1 and (
                    vrt_dst.colorinterp[indexes[0] - 1] is not ColorInterp.palette
                ):
                    tmp_dst.colorinterp = [ColorInterp.gray]
                else:
                    tmp_dst.colorinterp = [vrt_dst.colorinterp[b - 1] for b in indexes]

                if colormap:
                    if tmp_dst.colorinterp[0] is not ColorInterp.palette:
                        tmp_dst.colorinterp = [ColorInterp.palette]
                        warnings.warn(
                            "Dataset color interpretation was set to `Palette`"
                        )
                    tmp_dst.write_colormap(1, colormap)

                elif tmp_dst.colorinterp[0] is ColorInterp.palette:
                    try:
                        tmp_dst.write_colormap(1, vrt_dst.colormap(1))
                    except ValueError:
                        warnings.warn(
                            "Dataset has `Palette` color interpretation"
                            " but is missing colormap information"
                        )

                wind = list(tmp_dst.block_windows(1))

                if not quiet:
                    click.echo("Reading input: {}".format(source), err=True)

                fout = ctx.enter_context(open(os.devnull, "w")) if quiet else sys.stderr
                if quiet is False and progress_out:
                    fout = progress_out

                with click.progressbar(wind, file=fout, show_percent=True) as windows:  # type: ignore
                    for _, w in windows:
                        matrix = vrt_dst.read(window=w, indexes=indexes)
                        tmp_dst.write(matrix, window=w)

                        if add_mask or mask:
                            # Cast mask to uint8 to fix rasterio 1.1.2 error (ref #115)
                            mask_value = vrt_dst.dataset_mask(window=w).astype("uint8")
                            tmp_dst.write_mask(mask_value, window=w)

                if overview_level is None:
                    overview_level = get_maximum_overview_level(
                        vrt_dst.width, vrt_dst.height, minsize=tilesize
                    )

                if not quiet and overview_level:
                    click.echo("Adding overviews...", err=True)

                overviews = [decimation_base**j for j in range(1, overview_level + 1)]
                tmp_dst.build_overviews(overviews, ResamplingEnums[overview_resampling])

                if not quiet:
                    click.echo("Updating dataset tags...", err=True)

                for i, b in enumerate(src_indexes):
                    tmp_dst.set_band_description(i + 1, src_dst.descriptions[b - 1])
                    if forward_band_tags:
                        tmp_dst.update_tags(bidx=i + 1, **src_dst.tags(b))

                tags = src_dst.tags()
                tags.update(
                    {
                        "OVERVIEW_RESAMPLING": ResamplingEnums[
                            overview_resampling
                        ].name.upper()
                    }
                )
                if tms:
                    default_zoom = tms.zoom_for_res(
                        max(tmp_dst.res),
                        max_z=30,
                        zoom_level_strategy=zoom_level_strategy,
                    )
                    tags.update(
                        {
                            "TILING_SCHEME_NAME": tms.id or "CUSTOM",
                            "TILING_SCHEME_ZOOM_LEVEL": (
                                zoom_level if zoom_level is not None else default_zoom
                            ),
                        }
                    )
                    if aligned_levels:
                        tags["TILING_SCHEME_ALIGNED_LEVELS"] = aligned_levels

                if additional_cog_metadata:
                    tags.update(**additional_cog_metadata)

                if forward_ns_tags:
                    namespaces = src_dst.tag_namespaces()
                    for ns in namespaces:
                        if ns in ["DERIVED_SUBDATASETS", "IMAGE_STRUCTURE"]:
                            continue
                        tmp_dst.update_tags(ns=ns, **src_dst.tags(ns=ns))

                tmp_dst.update_tags(**tags)
                tmp_dst._set_all_scales([vrt_dst.scales[b - 1] for b in indexes])
                tmp_dst._set_all_offsets([vrt_dst.offsets[b - 1] for b in indexes])

                if not quiet:
                    click.echo("Writing output to: {}".format(dst_path), err=True)

                if use_cog_driver:
                    if not GDALVersion.runtime().at_least("3.1"):
                        raise Exception(
                            "GDAL 3.1 or above required to use the COG driver."
                        )

                    dst_kwargs["driver"] = "COG"

                    if add_mask and dst_kwargs.get("compress", "") != "JPEG":
                        warnings.warn(
                            "With GDAL COG driver, mask band will be translated to an alpha band."
                        )
                    if overview_level == 0:
                        dst_kwargs["overviews"] = "NONE"
                    dst_kwargs["overview_resampling"] = overview_resampling
                    dst_kwargs["warp_resampling"] = resampling
                    dst_kwargs["blocksize"] = tilesize
                    dst_kwargs.pop("blockxsize", None)
                    dst_kwargs.pop("blockysize", None)
                    dst_kwargs.pop("tiled", None)
                    dst_kwargs.pop("photometric", None)

                    copy(tmp_dst, dst_path, **dst_kwargs)

                else:
                    copy(tmp_dst, dst_path, copy_src_overviews=True, **dst_kwargs)


def cog_validate(  # noqa: C901
    src_path: Union[str, pathlib.PurePath],
    strict: bool = False,
    config: Optional[Dict] = None,
    quiet: bool = False,
) -> Tuple[bool, List[str], List[str]]:
    """
    Validate Cloud Optimized Geotiff.

    This script is the rasterio equivalent of
    https://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/validate_cloud_optimized_geotiff.py

    Parameters
    ----------
    src_path: str or PathLike object
        A dataset path or URL. Will be opened in "r" mode.
    strict: bool
        Treat warnings as errors
    quiet: bool
        Remove standard outputs

    Returns
    -------
    is_valid: bool
        True is src_path is a valid COG.
    errors: list
        List of validation errors.
    warnings: list
        List of validation warnings.

    """
    errors: List[str] = []
    warnings: List[str] = []
    details: Dict[str, Any] = {}

    config = config or {}

    if not GDALVersion.runtime().at_least("2.2"):
        raise Exception("GDAL 2.2 or above required")

    with rasterio.Env(**config):
        with rasterio.open(src_path) as src:
            if not src.driver == "GTiff":
                errors.append("The file is not a GeoTIFF")
                if not quiet:
                    click.secho("The following errors were found:", fg="red", err=True)
                    for e in errors:
                        click.echo("- " + e, err=True)

                return False, errors, warnings

            if any(pathlib.Path(x).suffix.lower() == ".ovr" for x in src.files):
                errors.append(
                    "Overviews found in external .ovr file. They should be internal"
                )

            overviews = src.overviews(1)
            if src.width > 512 and src.height > 512:
                if src.block_shapes and src.block_shapes[0][1] == src.width:
                    errors.append(
                        "The file is greater than 512xH or 512xW, but is not tiled"
                    )

                if not overviews:
                    warnings.append(
                        "The file is greater than 512xH or 512xW, it is recommended "
                        "to include internal overviews"
                    )

            ifd_offset = int(src.get_tag_item("IFD_OFFSET", "TIFF", bidx=1))
            # Starting from GDAL 3.1, GeoTIFF and COG have ghost headers
            # e.g:
            # """
            # GDAL_STRUCTURAL_METADATA_SIZE=000140 bytes
            # LAYOUT=IFDS_BEFORE_DATA
            # BLOCK_ORDER=ROW_MAJOR
            # BLOCK_LEADER=SIZE_AS_UINT4
            # BLOCK_TRAILER=LAST_4_BYTES_REPEATED
            # KNOWN_INCOMPATIBLE_EDITION=NO
            # """
            #
            # This header should be < 200bytes
            if ifd_offset > 300:
                errors.append(
                    f"The offset of the main IFD should be < 300. It is {ifd_offset} instead"
                )

            ifd_offsets = [ifd_offset]
            details["ifd_offsets"] = {}
            details["ifd_offsets"]["main"] = ifd_offset

            # Optimizations are usually invalidated when a COG is modified.
            # When this happens, GDAL >= 3.1 flags the modifications by updating the ghost headers.
            gdal_ghost_headers = src.get_tag_item("GDAL_STRUCTURAL_METADATA", "TIFF")
            if (
                gdal_ghost_headers is not None
                and "KNOWN_INCOMPATIBLE_EDITION=YES" in gdal_ghost_headers
            ):
                errors.append(
                    "This file used to have optimizations in its layout, "
                    "but those have been, at least partly, invalidated by later changes"
                )

            if overviews and overviews != sorted(overviews):
                errors.append("Overviews should be sorted")

            for ix, dec in enumerate(overviews):

                # NOTE: Size check is handled in rasterio `src.overviews` methods
                # https://github.com/mapbox/rasterio/blob/4ebdaa08cdcc65b141ed3fe95cf8bbdd9117bc0b/rasterio/_base.pyx
                # We just need to make sure the decimation level is > 1
                if not dec > 1:
                    errors.append(
                        "Invalid Decimation {} for overview level {}".format(dec, ix)
                    )

                # Check that the IFD of descending overviews are sorted by increasing
                # offsets
                ifd_offset = int(src.get_tag_item("IFD_OFFSET", "TIFF", bidx=1, ovr=ix))
                ifd_offsets.append(ifd_offset)

                details["ifd_offsets"]["overview_{}".format(ix)] = ifd_offset
                if ifd_offsets[-1] < ifd_offsets[-2]:
                    if ix == 0:
                        errors.append(
                            "The offset of the IFD for overview of index {} is {}, "
                            "whereas it should be greater than the one of the main "
                            "image, which is at byte {}".format(
                                ix, ifd_offsets[-1], ifd_offsets[-2]
                            )
                        )
                    else:
                        errors.append(
                            "The offset of the IFD for overview of index {} is {}, "
                            "whereas it should be greater than the one of index {}, "
                            "which is at byte {}".format(
                                ix, ifd_offsets[-1], ix - 1, ifd_offsets[-2]
                            )
                        )

            # Get blocks size
            block_size = src.block_shapes[0]

            # Extract number of blocks per row and column
            yblocks = (src.height + block_size[1] - 1) // block_size[1]
            xblocks = (src.width + block_size[0] - 1) // block_size[0]

            data_offsets = []
            details["data_offsets"] = {}

            # Find the first block with a valid block_offset
            for y in range(yblocks):
                for x in range(xblocks):
                    data_offset = get_data_offset(src, x, y)
                    if data_offset > 0:
                        data_offsets = [data_offset]
                        details["data_offsets"] = {}
                        details["data_offsets"]["main"] = data_offset
                        break
                else:
                    continue
                break

            for ix, _dec in enumerate(overviews):

                # Get the width and height of the overview
                overview_width = src.width // (_dec)
                overview_height = src.height // (_dec)

                # Extract number of blocks per row and column
                yblocks = (overview_height + block_size[1] - 1) // block_size[1]
                xblocks = (overview_width + block_size[0] - 1) // block_size[0]

                for y in range(yblocks):
                    for x in range(xblocks):
                        data_offset = get_data_offset(src, x, y, ix)
                        if data_offset > 0:
                            data_offset = int(data_offset) if data_offset else 0
                            data_offsets.append(data_offset)
                            details["data_offsets"][
                                "overview_{}".format(ix)
                            ] = data_offset
                            break
                    else:
                        continue
                    break

            if data_offsets[-1] != 0 and data_offsets[-1] < ifd_offsets[-1]:
                if len(overviews) > 0:
                    errors.append(
                        "The offset of the first block of the smallest overview "
                        "should be after its IFD"
                    )
                else:
                    errors.append(
                        "The offset of the first block of the image should "
                        "be after its IFD"
                    )

            for i in range(len(data_offsets) - 2, 0, -1):
                if data_offsets[i] < data_offsets[i + 1]:
                    errors.append(
                        "The offset of the first block of overview of index {} should "
                        "be after the one of the overview of index {}".format(i - 1, i)
                    )

            if len(data_offsets) >= 2 and data_offsets[0] < data_offsets[1]:
                errors.append(
                    "The offset of the first block of the main resolution image "
                    "should be after the one of the overview of index {}".format(
                        len(overviews) - 1
                    )
                )

        for ix, _dec in enumerate(overviews):
            with rasterio.open(src_path, OVERVIEW_LEVEL=ix) as ovr_dst:
                if ovr_dst.width > 512 and ovr_dst.height > 512:
                    if (
                        ovr_dst.block_shapes
                        and ovr_dst.block_shapes[0][1] == ovr_dst.width
                    ):
                        errors.append("Overview of index {} is not tiled".format(ix))

    if warnings and not quiet:
        click.secho("The following warnings were found:", fg="yellow", err=True)
        for w in warnings:
            click.echo("- " + w, err=True)
        click.echo(err=True)

    if errors and not quiet:
        click.secho("The following errors were found:", fg="red", err=True)
        for e in errors:
            click.echo("- " + e, err=True)

    is_valid = False if errors or (warnings and strict) else True

    return is_valid, errors, warnings


def get_data_offset(src, x, y, ix=None):
    """
    Retrieves the data offset from the source based on the given block coordinates.

    Parameters:
        src : object
            The source from which to retrieve the data offset.
        x : int
            The x-coordinate of the block.
        y : int
            The y-coordinate of the block.
        ix : int, optional
            The index of the overview. Default is None.

    Returns:
        int
            The data offset retrieved from the source. Returns 0 if the data offset is None.
    """
    data_offset = src.get_tag_item(
        "BLOCK_OFFSET_%d_%d" % (x, y), "TIFF", bidx=1, ovr=ix
    )
    return int(data_offset) if data_offset is not None else 0


def cog_info(
    src_path: Union[str, pathlib.PurePath],
    strict: bool = False,
    config: Optional[Dict] = None,
    quiet: bool = True,
) -> models.Info:
    """Get general info and validate Cloud Optimized Geotiff."""
    config = config or {}

    is_valid, validation_errors, validation_warnings = cog_validate(
        src_path, strict=strict, config=config, quiet=quiet
    )

    with rasterio.Env(**config):
        with rasterio.open(src_path) as src_dst:
            driver = src_dst.driver
            compression = getattr(src_dst.compression, "value", src_dst.compression)
            colorspace = src_dst.photometric.value if src_dst.photometric else None
            overviews = src_dst.overviews(1)

            tags = {"Image Metadata": src_dst.tags()}
            namespaces = src_dst.tag_namespaces()
            for ns in namespaces:
                if ns in ["DERIVED_SUBDATASETS"]:
                    continue
                tags.update({str.title(ns).replace("_", " "): src_dst.tags(ns=ns)})

            band_metadata = {
                f"Band {ix}": models.BandMetadata(
                    **{
                        "Description": src_dst.descriptions[ix - 1],
                        "ColorInterp": src_dst.colorinterp[ix - 1].name,
                        "Offset": src_dst.offsets[ix - 1],
                        "Scale": src_dst.scales[ix - 1],
                        "Metadata": src_dst.tags(ix),
                    }
                )
                for ix in src_dst.indexes
            }

            try:
                colormap = src_dst.colormap(1)
            except ValueError:
                colormap = None

            profile = models.Profile(
                Bands=src_dst.count,
                Width=src_dst.width,
                Height=src_dst.height,
                Tiled=(src_dst.block_shapes[0][1] != src_dst.width),
                Dtype=src_dst.dtypes[0],
                Interleave=(
                    src_dst.interleaving.value if src_dst.interleaving else "UNKNOWN"
                ),
                AlphaBand=utils.has_alpha_band(src_dst),
                InternalMask=utils.has_mask_band(src_dst),
                Nodata=src_dst.nodata,
                ColorInterp=tuple([color.name for color in src_dst.colorinterp]),
                ColorMap=colormap is not None,
                Scales=src_dst.scales,
                Offsets=src_dst.offsets,
            )

            try:
                crs = (
                    f"EPSG:{src_dst.crs.to_epsg()}"
                    if src_dst.crs.to_epsg()
                    else src_dst.crs.to_wkt()
                )
            except AttributeError:
                crs = None

            minzoom: Optional[int] = None
            maxzoom: Optional[int] = None
            try:
                minzoom, maxzoom = utils.get_zooms(src_dst)
            except Exception:
                pass

            geo = models.Geo(
                CRS=crs,
                BoundingBox=tuple(src_dst.bounds),
                Origin=(src_dst.transform.c, src_dst.transform.f),
                Resolution=(src_dst.transform.a, src_dst.transform.e),
                MinZoom=minzoom,
                MaxZoom=maxzoom,
            )

            ifds = [
                models.IFD(
                    Level=0,
                    Width=src_dst.width,
                    Height=src_dst.height,
                    Blocksize=src_dst.block_shapes[0],
                    Decimation=0,
                )
            ]

        for ix, decim in enumerate(overviews):
            with rasterio.open(src_path, OVERVIEW_LEVEL=ix) as ovr_dst:
                ifds.append(
                    models.IFD(
                        Level=ix + 1,
                        Width=ovr_dst.width,
                        Height=ovr_dst.height,
                        Blocksize=ovr_dst.block_shapes[0],
                        Decimation=decim,
                    )
                )

    return models.Info(
        Path=str(src_path),
        Driver=driver,
        COG=is_valid,
        Compression=compression,
        ColorSpace=colorspace,
        COG_errors=validation_errors or None,
        COG_warnings=validation_warnings or None,
        Profile=profile,
        GEO=geo,
        Tags=tags,
        Band_Metadata=band_metadata,  # type: ignore [call-arg]
        IFD=ifds,
    )



"""Rio_cogeo.scripts.cli."""

import json
import os
import typing
from typing import Optional

import click
import morecantile
import numpy
from morecantile import TileMatrixSet
from rasterio.rio import options

from rio_cogeo import __version__ as cogeo_version
from rio_cogeo.cogeo import (
    RIOResampling,
    WarpResampling,
    cog_info,
    cog_translate,
    cog_validate,
)
from rio_cogeo.profiles import cog_profiles

IN_MEMORY_THRESHOLD = int(os.environ.get("IN_MEMORY_THRESHOLD", 10980 * 10980))


class BdxParamType(click.ParamType):
    """Band index type."""

    name = "bidx"

    def convert(self, value, param, ctx):
        """Validate and parse band index."""
        try:
            bands = [int(x) for x in value.split(",")]
            assert all(b > 0 for b in bands)
            return bands

        except (ValueError, AttributeError, AssertionError) as e:
            raise click.ClickException(
                "bidx must be a string of comma-separated integers (> 0), "
                "representing the band indexes."
            ) from e


class NodataParamType(click.ParamType):
    """Nodata type."""

    name = "nodata"

    def convert(self, value, param, ctx):
        """Validate and parse band index."""
        try:
            if value.lower() == "nan":
                return numpy.nan
            elif value.lower() in ["nil", "none", "nada"]:
                return None
            else:
                return float(value)
        except (TypeError, ValueError) as e:
            raise click.ClickException(
                "{} is not a valid nodata value.".format(value)
            ) from e


class ThreadsParamType(click.ParamType):
    """num_threads index type."""

    name = "threads"

    def convert(self, value, param, ctx):
        """Validate and parse thread number."""
        try:
            if value.lower() == "all_cpus":
                return "ALL_CPUS"
            else:
                return int(value)
        except (TypeError, ValueError) as e:
            raise click.ClickException(
                "{} is not a valid thread value.".format(value)
            ) from e


@click.group(short_help="Create and Validate COGEO")
@click.version_option(version=cogeo_version, message="%(version)s")
def cogeo():
    """Rasterio cogeo subcommands."""
    pass


@cogeo.command(short_help="Create COGEO")
@options.file_in_arg
@options.file_out_arg
@click.option("--bidx", "-b", type=BdxParamType(), help="Band indexes to copy.")
@click.option(
    "--cog-profile",
    "-p",
    "cogeo_profile",
    type=click.Choice(cog_profiles.keys(), case_sensitive=False),
    default="deflate",
    help="CloudOptimized GeoTIFF profile.",
    show_default=True,
)
@click.option(
    "--nodata",
    type=NodataParamType(),
    metavar="NUMBER|nan",
    help="Set nodata masking values for input dataset.",
)
@click.option(
    "--add-mask",
    is_flag=True,
    help="Force output dataset creation with an internal mask (convert alpha "
    "band or nodata to mask).",
)
@click.option("--blocksize", type=int, help="Overwrite profile's tile size.")
@options.dtype_opt
@click.option(
    "--overview-level",
    type=int,
    help="Overview level (if not provided, appropriate overview level will be "
    "selected until the smallest overview is smaller than the value of the "
    "internal blocksize)",
)
@click.option(
    "--decimation-base",
    type=int,
    help="Decimation base, how to sub-divide a raster for each overview level. "
    "Also requires that --overview-level is provided.",
    default=2,
    show_default=True,
)
@click.option(
    "--overview-resampling",
    help="Overview creation resampling algorithm.",
    type=click.Choice(list(typing.get_args(RIOResampling))),
    default="nearest",
    show_default=True,
)
@click.option(
    "--overview-blocksize",
    help="Overview's internal tile size (default can be defined by "
    "GDAL_TIFF_OVR_BLOCKSIZE env). The default is 128, or starting with GDAL 3.1 to use the same block size as the full-resolution dataset if possible).",
)
@click.option(
    "--web-optimized", "-w", is_flag=True, help="Create COGEO optimized for Web."
)
@click.option(
    "--zoom-level-strategy",
    type=click.Choice(["lower", "upper", "auto"], case_sensitive=False),
    default="auto",
    help="Strategy to determine zoom level.",
    show_default=True,
)
@click.option(
    "--zoom-level",
    type=int,
    help="Zoom level number for the highest resolution. If this option is specified, `--zoom-level-strategy` is ignored.",
)
@click.option(
    "--aligned-levels",
    type=int,
    help="Number of overview levels for which GeoTIFF tile and tiles defined in the tiling scheme match.",
)
@click.option(
    "--resampling",
    "-r",
    help="Resampling algorithm. Will only be applied with the `--web-optimized` option.",
    type=click.Choice(list(typing.get_args(WarpResampling))),
    default="nearest",
    show_default=True,
)
@click.option(
    "--in-memory/--no-in-memory",
    default=None,
    help="Force processing raster in memory / not in memory (default: process in memory "
    "if smaller than {:.0f} million pixels)".format(IN_MEMORY_THRESHOLD // 1e6),
)
@click.option(
    "--allow-intermediate-compression",
    default=False,
    is_flag=True,
    help="Allow intermediate file compression to reduce memory/disk footprint.",
    show_default=True,
)
@click.option(
    "--forward-band-tags",
    default=False,
    is_flag=True,
    help="Forward band tags to output bands.",
    show_default=True,
)
@click.option(
    "--forward-ns-tags",
    default=False,
    is_flag=True,
    help="Forward namespaced tags to output dataset.",
    show_default=True,
)
@click.option(
    "--threads",
    type=ThreadsParamType(),
    default="ALL_CPUS",
    help="Number of worker threads for multi-threaded compression.",
    show_default=True,
)
@click.option(
    "--use-cog-driver",
    help="Use GDAL COG Driver (require GDAL>=3.1).",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--tms",
    help="Path to TileMatrixSet JSON file.",
    type=click.Path(),
)
@options.creation_options
@click.option(
    "--config",
    "config",
    metavar="NAME=VALUE",
    multiple=True,
    callback=options._cb_key_val,
    help="GDAL configuration options.",
)
@click.option(
    "--quiet", "-q", help="Remove progressbar and other non-error output.", is_flag=True
)
def create(
    input,
    output,
    bidx,
    cogeo_profile,
    nodata,
    dtype,
    add_mask,
    blocksize,
    overview_level,
    decimation_base,
    overview_resampling,
    overview_blocksize,
    web_optimized,
    zoom_level_strategy,
    zoom_level,
    aligned_levels,
    resampling,
    in_memory,
    allow_intermediate_compression,
    forward_band_tags,
    forward_ns_tags,
    threads,
    use_cog_driver,
    tms,
    creation_options,
    config,
    quiet,
):
    """Create Cloud Optimized Geotiff."""
    config.update(
        {
            "GDAL_NUM_THREADS": threads,
            "GDAL_TIFF_INTERNAL_MASK": os.environ.get("GDAL_TIFF_INTERNAL_MASK", True),
        }
    )

    output_profile = cog_profiles.get(cogeo_profile)
    output_profile.update({"BIGTIFF": os.environ.get("BIGTIFF", "IF_SAFER")})
    if creation_options:
        output_profile.update(creation_options)

    tilematrixset: Optional[TileMatrixSet] = None
    if tms:
        with open(tms, "r") as f:
            tilematrixset = TileMatrixSet(**json.load(f))

    elif web_optimized:
        tilematrixset = morecantile.tms.get("WebMercatorQuad")

    if tilematrixset:
        if not blocksize:
            click.secho(
                f"Defining `blocksize` from {tilematrixset.id} TileMatrixSet `tileWidth` and `tileHeight`",
                fg="yellow",
            )

            blocksize = min(
                tilematrixset.matrix(tilematrixset.minzoom).tileHeight,
                tilematrixset.matrix(tilematrixset.minzoom).tileWidth,
            )

        if not overview_blocksize:
            click.secho(
                f"Defining overview's `blocksize` to match the high resolution `blocksize`: {blocksize}",
                fg="yellow",
            )
            overview_blocksize = blocksize

    if blocksize:
        output_profile["blockxsize"] = blocksize
        output_profile["blockysize"] = blocksize

    overview_blocksize = overview_blocksize or os.environ.get("GDAL_TIFF_OVR_BLOCKSIZE")
    if overview_blocksize:
        config.update({"GDAL_TIFF_OVR_BLOCKSIZE": str(overview_blocksize)})

    cog_translate(
        input,
        output,
        output_profile,
        indexes=bidx,
        nodata=nodata,
        dtype=dtype,
        add_mask=add_mask,
        overview_level=overview_level,
        overview_resampling=overview_resampling,
        zoom_level_strategy=zoom_level_strategy,
        zoom_level=zoom_level,
        aligned_levels=aligned_levels,
        resampling=resampling,
        in_memory=in_memory,
        config=config,
        allow_intermediate_compression=allow_intermediate_compression,
        forward_band_tags=forward_band_tags,
        forward_ns_tags=forward_ns_tags,
        tms=tilematrixset,
        use_cog_driver=use_cog_driver,
        quiet=quiet,
        decimation_base=decimation_base,
    )


@cogeo.command(short_help="Validate COGEO")
@options.file_in_arg
@click.option(
    "--strict",
    default=False,
    is_flag=True,
    help="Treat warnings as errors.",
    show_default=True,
)
@click.option(
    "--config",
    "config",
    metavar="NAME=VALUE",
    multiple=True,
    callback=options._cb_key_val,
    help="GDAL configuration options.",
)
def validate(input, strict, config):
    """Validate Cloud Optimized Geotiff."""
    is_valid, _, _ = cog_validate(input, strict=strict, config=config)
    if is_valid:
        click.echo("{} is a valid cloud optimized GeoTIFF".format(input))
    else:
        click.echo("{} is NOT a valid cloud optimized GeoTIFF".format(input))


@cogeo.command(short_help="Lists information about a raster dataset.")
@options.file_in_arg
@click.option(
    "--json",
    "to_json",
    default=False,
    is_flag=True,
    help="Print as JSON.",
    show_default=True,
)
@click.option(
    "--config",
    "config",
    metavar="NAME=VALUE",
    multiple=True,
    callback=options._cb_key_val,
    help="GDAL configuration options.",
)
def info(input, to_json, config):  # noqa: C901
    """Dataset info."""
    metadata = cog_info(input, config=config)

    if to_json:
        click.echo(metadata.model_dump_json(exclude_none=True, by_alias=True))
    else:

        sep = 25
        click.echo(
            f"""{click.style('Driver:', bold=True)} {metadata.Driver}
{click.style('File:', bold=True)} {metadata.Path}
{click.style('COG:', bold=True)} {metadata.COG}
{click.style('Compression:', bold=True)} {metadata.Compression}
{click.style('ColorSpace:', bold=True)} {metadata.ColorSpace}

{click.style('Profile', bold=True)}
    {click.style("Width:", bold=True):<{sep}} {metadata.Profile.Width}
    {click.style("Height:", bold=True):<{sep}} {metadata.Profile.Height}
    {click.style("Bands:", bold=True):<{sep}} {metadata.Profile.Bands}
    {click.style("Tiled:", bold=True):<{sep}} {metadata.Profile.Tiled}
    {click.style("Dtype:", bold=True):<{sep}} {metadata.Profile.Dtype}
    {click.style("NoData:", bold=True):<{sep}} {metadata.Profile.Nodata}
    {click.style("Alpha Band:", bold=True):<{sep}} {metadata.Profile.AlphaBand}
    {click.style("Internal Mask:", bold=True):<{sep}} {metadata.Profile.InternalMask}
    {click.style("Interleave:", bold=True):<{sep}} {metadata.Profile.Interleave}
    {click.style("ColorMap:", bold=True):<{sep}} {metadata.Profile.ColorMap}
    {click.style("ColorInterp:", bold=True):<{sep}} {metadata.Profile.ColorInterp}
    {click.style("Scales:", bold=True):<{sep}} {metadata.Profile.Scales}
    {click.style("Offsets:", bold=True):<{sep}} {metadata.Profile.Offsets}"""
        )

        click.echo(
            f"""
{click.style('Geo', bold=True)}
    {click.style("Crs:", bold=True):<{sep}} {metadata.GEO.CRS}
    {click.style("Origin:", bold=True):<{sep}} {metadata.GEO.Origin}
    {click.style("Resolution:", bold=True):<{sep}} {metadata.GEO.Resolution}
    {click.style("BoundingBox:", bold=True):<{sep}} {metadata.GEO.BoundingBox}
    {click.style("MinZoom:", bold=True):<{sep}} {metadata.GEO.MinZoom}
    {click.style("MaxZoom:", bold=True):<{sep}} {metadata.GEO.MaxZoom}"""
        )

        for ns, values in metadata.Tags.items():
            click.echo(
                f"""
{click.style(ns, bold=True)}"""
            )
            for key, val in values.items():
                click.echo(
                    f"""    {click.style(key, underline=True, bold=True)}: {val}"""
                )

        for ns, meta in metadata.Band_Metadata.items():
            click.echo(
                f"""
{click.style(ns, bold=True)}"""
            )

            if meta.Description:
                click.echo(
                    f"""    {click.style("Description", underline=True, bold=True)}: {meta.Description}"""
                )
            click.echo(
                f"""    {click.style("ColorInterp", underline=True, bold=True)}: {meta.ColorInterp}"""
            )
            if meta.Offset != 0.0 and meta.Scale != 1.0:
                click.echo(
                    f"""    {click.style("Offset", underline=True, bold=True)}: {meta.Offset}"""
                )
                click.echo(
                    f"""    {click.style("Scale", underline=True, bold=True)}: {meta.Scale}"""
                )

            if meta.Metadata:
                click.echo(
                    f"""    {click.style("Metadata", underline=True, bold=True)}:"""
                )
                for key, val in meta.Metadata.items():
                    click.echo(
                        f"""        {click.style(key, underline=True, bold=True)}: {val}"""
                    )

        click.echo(
            f"""
{click.style('IFD', bold=True)}
    {click.style('Id', underline=True, bold=True):<20}{click.style('Size', underline=True, bold=True):<27}{click.style('BlockSize', underline=True, bold=True):<26}{click.style('Decimation', underline=True, bold=True)}"""
        )

        for ifd in metadata.IFD:
            wh = f"{ifd.Width}x{ifd.Height}"
            bl = f"{ifd.Blocksize[1]}x{ifd.Blocksize[0]}"
            click.echo(f"""    {ifd.Level:<8}{wh:<15}{bl:<14}{ifd.Decimation}""")

        if metadata.COG_errors or metadata.COG_warnings:
            click.echo(
                f"""
{click.style('COG Validation info', bold=True)}"""
            )
            for error in metadata.COG_errors or []:
                click.secho(f"""    - {error} (error)""", fg="red")
            for warning in metadata.COG_warnings or []:
                click.secho(f"""    - {warning} (warning)""", fg="yellow")



"""rio_cogeo: Cloud Optimized GeoTIFF creation and validation plugin for rasterio."""

__version__ = "7.0.2"

from .cogeo import cog_info, cog_translate, cog_validate  # noqa
from .profiles import cog_profiles  # noqa
